#!/usr/bin/env python3
"""Display sparklines from a sequence of numbers.
"""
from decimal import Decimal
from math import inf
from typing import Any, List, Optional, Union

from typing_extensions import TypeGuard

from milc import cli

spark_chars = '▁▂▃▄▅▆▇█'


def is_number(i: Any) -> TypeGuard[Union[int, float, Decimal]]:
    """Returns true if i is a number. Used to filter non-numbers from a list.
    """
    return isinstance(i, (int, float, Decimal))


def sparkline(
    number_list: List[Optional[Union[int, float, Decimal]]],
    *,
    min_value: Optional[Union[int, float, Decimal]] = None,
    max_value: Optional[Union[int, float, Decimal]] = None,
    highlight_low: float = -inf,
    highlight_high: float = inf,
    highlight_low_color: str = '',
    highlight_high_color: str = '',
    negative_color: str = '{fg_red}',
    positive_color: str = '',
    highlight_low_reset: str = '{fg_reset}',
    highlight_high_reset: str = '{fg_reset}',
    negative_reset: str = '{fg_reset}',
    positive_reset: str = '{fg_reset}',
) -> str:
    """Display a sparkline from a sequence of numbers.

    If you wish to exclude extreme values, or you want to limit the set of characters used, you can adjust `min_value` and `max_value` to your own values. Values between your actual min/max will exclude datapoints, while values outside your actual min/max will compress your data into fewer sparks.

    If you want to highlight data that is too low or too high you can use 'highlight_low' and `highlight_high` to set this. You will also need to set your colors, see below for more details.

    By default this function will display negative numbers in red and positive numbers in the system default color. You can use `negative_color`, `negative_reset`, `positive_color`, and `positive_reset` to change this behavior.

    If you wish to color your sparkline according to other rules it is recommended to generate it without color and then add color yourself.

    ### Arguments

        min_value
            The lowest value in your sparkline. If not provided it will be determined automatically.

        max_value
            The highest value in your sparkline. If not provided it will be determined automatically.

        highlight_low
            When a number is less than this value it will be highlighted with `highlight_low_color`.

        highlight_high
            When a number is greater than this value it will be highlighted with `highlight_high_color`.

        highlight_low_color
            A MILC or ANSI color code to apply to integers less than highlight_low.

        highlight_high_color
            A MILC or ANSI color code to apply to integers greater than highlight_high.

        negative_color
            A MILC or ANSI color code to apply to integers less than 0.

        positive_color
            A MILC or ANSI color code to apply to integers greater than 0.

        highlight_low_reset
            A MILC or ANSI color code to reset the color code applied in `highlight_low_color`. This is usually `{fg_reset}`, `{bg_reset}`, or `{style_reset_all}`.

        highlight_high_reset
            A MILC or ANSI color code to reset the color code applied in `highlight_high_color`. This is usually `{fg_reset}`, `{bg_reset}`, or `{style_reset_all}`.

        negative_reset
            A MILC or ANSI color code to reset the color code applied in `negative_color`. This is usually `{fg_reset}`, `{bg_reset}`, or `{style_reset_all}`.

        positive_reset
            A MILC or ANSI color code to reset the color code applied in `positive_color`. This is usually `{fg_reset}`, `{bg_reset}`, or `{style_reset_all}`.
    """
    if min_value is None:
        min_value = min(filter(is_number, number_list), default=None)

    if max_value is None:
        max_value = max(filter(is_number, number_list), default=None)

    if min_value is None or max_value is None:
        return ' ' * len(number_list)

    int_range = float(max_value) - float(min_value)
    sparks = []

    for i in number_list:
        # Handle non-numeric and out-of-bounds values
        if not is_number(i):
            sparks.append(' ')
            continue

        if i < min_value or i > max_value:
            cli.log.debug('Skipping out of bounds value %s', i)
            continue

        # Determine the bucket for this value
        if int_range == 0:
            spark_int = 0
        else:
            spark_int = int((float(i) - float(min_value)) / int_range * 8)

        spark_int = min(spark_int, 7)

        # Determine the color for this value
        color = positive_color
        reset = positive_reset

        if i < 0:
            color = negative_color
            reset = negative_reset

        if i < highlight_low:
            color = highlight_low_color
            reset = highlight_low_reset

        if i > highlight_high:
            color = highlight_high_color
            reset = highlight_high_reset

        if color == '':
            reset = ''

        # Add this spark to the list
        sparks.append(''.join((color, spark_chars[spark_int], reset)))

    return ''.join(sparks)



"""Unit tests for milc.questions."""
import milc.questions


def test_format_prompt_no_args():
    """Ensure prompts without format args are returned unchanged (no ValueError for %% in prompt)."""
    assert milc.questions._format_prompt('Are you 50% sure?', (), {}) == 'Are you 50% sure?'


def test_format_prompt_with_args():
    """Ensure positional args are applied to the prompt."""
    assert milc.questions._format_prompt('Hello %s!', ('world',), {}) == 'Hello world!'


def test_format_prompt_with_kwargs():
    """Ensure keyword args are applied to the prompt."""
    assert milc.questions._format_prompt('Hello %(name)s!', (), {'name': 'world'}) == 'Hello world!'



"""Unit tests for milc.configuration.Configuration.
"""
import milc.configuration


def configuration_setup():
    """Setup a simple milc.configuration.Configuration object for testing.
    """
    config = milc.configuration.Configuration()
    config['a']['a'] = True
    config['a']['a'] = True
    config['a']['b'] = False
    config['user']['d'] = 'arbitrary string'
    config['user']['d'] = 'arbitrary string'

    return config


def test_Configuration_dictionary():
    """Test Configuration dictionary access
    """
    config = configuration_setup()

    assert config['a']['a'] is True
    assert config['a']['b'] is False
    assert config['a']['c'] is None


def test_Configuration_dictionary_write():
    """Test that Configuration dictionary writes can be read as attributes.
    """
    config = milc.configuration.Configuration()
    config['a']['a'] = True
    assert config.a.a is True


def test_Configuration_attribute():
    """Test Configuration attribute access
    """
    config = configuration_setup()

    assert config.a.a is True
    assert config.a.b is False
    assert config.a.c is None


def test_Configuration_attribute_write():
    """Test that Configuration attribute writes can be read as a dictionary method.
    """
    config = milc.configuration.Configuration()
    config.a.a = True
    assert config['a']['a'] is True


def test_Configuration_iteration():
    """Make sure we can iterate over our stored volues.
    """
    config = configuration_setup()

    for section, section_dict in config.items():
        if section == 'a':
            for key, value in section_dict.items():
                if key == 'a':
                    assert value is True
                elif key == 'b':
                    assert value is False
                else:
                    raise KeyError('Unexpected key found in configuration: ' + section + '.' + key)
        elif section == 'user':
            for key, value in section_dict.items():
                if key == 'd':
                    assert value == 'arbitrary string'
                else:
                    raise KeyError('Unexpected key found in configuration: ' + section + '.' + key)
        else:
            raise KeyError('Unexpected section found in configuration: ' + section)


def test_Configuration_parent_fallback():
    """Make sure that we properly pull in the parent value when the local value doesn't exist.
    """
    config = configuration_setup()

    assert config.a.d == 'arbitrary string'



import os
import re
import sys
from tempfile import mkstemp

from .common import check_assert, check_command, check_returncode


def test_example():
    result = check_command(sys.executable, 'example', '-h')
    check_returncode(result)
    check_assert(result, '{config,dashed-hello,hello,goodbye,config-file}' in result.stdout)


def test_example_version():
    result = check_command(sys.executable, 'example', '--version')
    check_returncode(result)
    check_assert(result, re.match(r'[0-9]*\.[0-9]*\.[0-9]*', result.stdout))


def test_example_config():
    fd, tempfile = mkstemp()

    try:
        os.close(fd)

        # Check initial state
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'config')
        check_returncode(result)
        check_assert(result, result.stdout == '')

        # Set some values in the configuration
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'config', 'general.name=Test', 'user.comma=true')
        check_returncode(result)
        check_assert(result, 'general.name: World -> Test' in result.stdout)
        check_assert(result, 'user.comma: None -> true' in result.stdout)

        # Make sure we get them back
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'config')
        check_returncode(result)
        check_assert(result, result.stdout == 'general.name=Test\nhello.comma=True\nuser.comma=True\n')

    finally:
        os.remove(tempfile)


def test_numeric_arguments():
    fd, tempfile = mkstemp()

    try:
        os.close(fd)

        # Set the count
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'config', 'hello.count=2')
        check_returncode(result)
        check_assert(result, 'hello.count: 1 -> 2' in result.stdout)

        # Make sure we get 2 prints
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'hello')
        check_returncode(result)
        print(repr(result.stdout))
        check_assert(result, result.stdout == 'Hello, World!\nHello, World!\n')

        # Make sure we get 1 prints
        result = check_command(sys.executable, 'example', '--no-color', '--config-file', tempfile, 'hello', '--count', '1')
        check_returncode(result)
        print(repr(result.stdout))
        check_assert(result, result.stdout == 'Hello, World!\n')

    finally:
        os.remove(tempfile)


def test_example_hello():
    result = check_command(sys.executable, 'example', '--config-file', os.devnull, 'hello')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, World!\n')


def test_example_hello_help():
    result = check_command(sys.executable, 'example', '--config-file', os.devnull, 'hello', '--help')
    check_returncode(result)
    check_assert(result, '[--no-comma]' in result.stdout)
    check_assert(result, '[--print-help]' in result.stdout)
    check_assert(result, '[--print-usage]' in result.stdout)


def test_example_hello_no_color():
    result = check_command(sys.executable, 'example', '--no-color', '--config-file', os.devnull, 'hello')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, World!\n')


def test_example_hello_no_color_no_unicode():
    result = check_command(sys.executable, 'example', '--no-color', '--no-unicode', '--config-file', os.devnull, 'hello')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, World!\n')


def test_example_hello_no_color_no_comma():
    result = check_command(sys.executable, 'example', '--no-color', '--config-file', os.devnull, 'hello', '--no-comma')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello World!\n')


def test_example_hello_no_color_no_unicode_no_comma():
    result = check_command(sys.executable, 'example', '--no-color', '--no-unicode', '--config-file', os.devnull, 'hello', '--no-comma')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello World!\n')


def test_example_dashed_hello():
    result = check_command(sys.executable, 'example', '--config-file', os.devnull, 'dashed-hello')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, dashed-subcommand World!\n')


def test_example_dashed_hello_dashed_name():
    result = check_command(sys.executable, 'example', '--config-file', os.devnull, 'dashed-hello', '--dashed-name', 'Tester')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, dashed-subcommand Tester!\n')


def test_example_config_file():
    result = check_command(sys.executable, 'example', 'config-file')
    check_returncode(result)
    lines = result.stdout.split('\n')
    filename = lines[0].split('=', 1)[1]
    filedir = lines[1].split('=', 1)[1]
    check_assert(result, len(lines) == 3)
    check_assert(result, filename.endswith('example.ini'))
    check_assert(result, filename.startswith(filedir))



"""Unit tests for milc._sparkline.
"""
from milc import sparkline


def test_sparkline_simple():
    """Test a simple sparkline
    """
    assert sparkline([5, 9, 3, 15]) == '▂▅▁█'


def test_sparkline_corner1():
    """Test a common corner case
    """
    assert sparkline([0, 1, 19, 20]) == '▁▁██'


def test_sparkline_corner2():
    """Test a common corner case
    """
    assert sparkline([0, 999, 4000, 4999, 7000, 7999]) == '▁▁▅▅██'


def test_sparkline_negative():
    """Ensure that negative numbers are highlighted.
    """
    assert sparkline([-1, 0, 1]) == '{fg_red}▁{fg_reset}▅█'


def test_sparkline_negative_positive():
    """Ensure that negative and positive numbers are highlighted.
    """
    assert sparkline([-1, 0, 1], positive_color='{fg_green}') == '{fg_red}▁{fg_reset}{fg_green}▅{fg_reset}{fg_green}█{fg_reset}'


def test_sparkline_highlight_high():
    """Ensure that highlight high numbers are highlighted.
    """
    assert sparkline([0, 1, 2], highlight_high=1, highlight_high_color='{fg_magenta}') == '▁▅{fg_magenta}█{fg_reset}'


def test_sparkline_highlight_low():
    """Ensure that highlight low numbers are highlighted.
    """
    assert sparkline([0, 1, 2], highlight_low=1, highlight_low_color='{fg_magenta}') == '{fg_magenta}▁{fg_reset}▅█'


def test_whitespace():
    """Ensure that whitespace is properly inserted.
    """
    assert sparkline([0, None, 1]) == '▁ █'


def test_exclude():
    """Ensure that out-of-bounds values are excluded.
    """
    assert sparkline([-1, 0, None, 1, 2], min_value=0, max_value=1) == '▁ █'



import sys
from typing import Optional


def _in_argv(argument: str) -> bool:
    """Returns true if the argument is found is sys.argv.

    Since long options can be passed as either '--option value' or '--option=value' we need to check for both forms.
    """
    for arg in sys.argv:
        if arg.split('=')[0] == argument:
            return True

    return False


def _index_argv(argument: str) -> Optional[int]:
    """Returns the location of the argument in sys.argv, or None.

    Since long options can be passed as either '--option value' or '--option=value' we need to check for both forms.
    """
    for i, arg in enumerate(sys.argv):
        if arg.split('=')[0] == argument:
            return i

    return None



"""Test the few things inside the `milc/__init__.py` file.
"""
import semver

import milc

EMOJI_LEVELS = ('CRITICAL', 'ERROR', 'FATAL', 'WARN', 'WARNING', 'INFO', 'DEBUG', 'NOTSET')


def test_version():
    """Make sure that version matches semver.
    """
    assert semver.VersionInfo.is_valid(milc.__VERSION__)


def test_emojis():
    """Make sure the emoji list is populated.
    """
    for emoji in milc.EMOJI_LOGLEVELS:
        if emoji not in EMOJI_LEVELS:
            raise KeyError(emoji)


def test_cli():
    """Make sure the cli is available.
    """
    assert milc.cli



"""Public interface for MILC.

This is where the public interface for `cli` is kept. This allows us to reinstantiate MILC without having to recreate the cli object, as well as allowing us to have a well defined public API.
"""
import sys
import warnings
from logging import Logger
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, Dict, Optional, Sequence, Type, TypeVar, Union

from halo import Halo
from typing_extensions import ParamSpec

from .attrdict import AttrDict
from .configuration import Configuration
from .milc import MILC

P = ParamSpec("P")
R = TypeVar("R")


class MILCInterface:
    def __init__(self) -> None:
        self._milc: Optional[MILC] = None
        self._name: Optional[str] = None
        self._author: Optional[str] = None
        self._version: Optional[str] = None
        self._logger: Optional[Logger] = None

    def milc_options(self, *, name: Optional[str] = None, author: Optional[str] = None, version: Optional[str] = None, logger: Optional[Logger] = None) -> None:
        if self._milc and self._milc._inside_context_manager:
            raise RuntimeError('You must run cli.milc_options() before cli() or anything else!')

        self._name = name or self._name
        self._author = author or self._author
        self._version = version or self._version
        self._logger = logger or self._logger
        self._milc = MILC(self._name, self._author, self._version, self._logger)

    @property
    def milc(self) -> MILC:
        if not self._milc:
            self._milc = MILC()

        return self._milc

    @property
    def args(self) -> AttrDict:
        return self.milc.args

    @property
    def config(self) -> Configuration:
        return self.milc.config

    @property
    def config_dir(self) -> Path:
        return self.milc.config_dir

    @property
    def config_source(self) -> Configuration:
        return self.milc.config_source

    @property
    def description(self) -> Optional[str]:
        return self.milc.description

    @property
    def interactive(self) -> bool:
        return self.milc.interactive

    @property
    def log(self) -> Logger:
        return self.milc.log

    @property
    def platform(self) -> str:
        return self.milc.platform

    @property
    def _subcommand(self) -> Any:
        warnings.warn("cli._subcommand has been deprecated, please use cli.subcommand_name to get the subcommand name instead.", stacklevel=2)

        return self.milc._subcommand

    @property
    def subcommands(self) -> Dict[str, Any]:
        return self.milc.subcommands

    @property
    def subcommand_name(self) -> Optional[str]:
        return self.milc.subcommand_name

    def echo(self, text: str, *args: Any, **kwargs: Any) -> None:
        """Print colorized text to stdout.

        ANSI color strings (such as {fg_blue}) will be converted into ANSI
        escape sequences, and the ANSI reset sequence will be added to all
        strings.

        If *args or **kwargs are passed they will be used to %-format the strings.
        """
        return self.milc.echo(text, *args, **kwargs)

    def run(self, command: Sequence[str], capture_output: bool = True, combined_output: bool = False, text: bool = True, **kwargs: Any) -> Any:  # FIXME: In python 3.10 we can use subprocess.CompletedProcess[bytes | str] instead
        """Run a command using `subprocess.run`, but using some different defaults.

        Unlike subprocess.run you must supply a sequence of arguments. You can use `shlex.split()` to build this from a string.

        The **kwargs arguments get passed directly to `subprocess.run`.

        Args:
            command
                A sequence where the first item is the command to run, and any remaining items are arguments to pass.

            capture_output
                Set to False to have output written to the terminal instead of being available in the returned `subprocess.CompletedProcess` instance.

            combined_output
                When true STDERR will be written to STDOUT. Equivalent to the shell construct `2>&1`.

            text
                Set to False to disable encoding and get `bytes()` from `.stdout` and `.stderr`.
        """
        return self.milc.run(command, capture_output, combined_output, text, **kwargs)

    def print_help(self, *args: Any, **kwargs: Any) -> None:
        """Print a help message for the main program or subcommand, depending on context.
        """
        return self.milc.print_help(*args, **kwargs)

    def print_usage(self, *args: Any, **kwargs: Any) -> None:
        """Print brief description of how the main program or subcommand is invoked, depending on context.
        """
        return self.milc.print_usage(*args, **kwargs)

    def add_argument(self, *args: Any, **kwargs: Any) -> None:
        """Wrapper to add arguments and track whether they were passed on the command line.
        """
        return self.milc.add_argument(*args, **kwargs)

    def acquire_lock(self, blocking: bool = True) -> bool:
        """Acquire the MILC lock for exclusive access to properties.
        """
        return self.milc.acquire_lock(blocking)

    def release_lock(self) -> None:
        """Release the MILC lock.
        """
        return self.milc.release_lock()

    def argument(self, *args: Any, **kwargs: Any) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator to add an argument to a MILC command or subcommand.
        """
        return self.milc.argument(*args, **kwargs)

    def save_config(self) -> None:
        """Save the current configuration to the config file.
        """
        return self.milc.save_config()

    def __call__(self) -> Any:
        """Execute the entrypoint function.
        """
        return self.milc()

    def entrypoint(self, description: str, deprecated: Optional[str] = None) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator that marks the entrypoint used when a subcommand is not supplied.
        Args:
            description
                A one-line description to display in --help

            deprecated
                Deprecation message. When set the subcommand will marked as deprecated and this message will be displayed in the help output.
        """
        return self.milc.entrypoint(description, deprecated)

    def subcommand(self, description: str, hidden: bool = False, **kwargs: Any) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator to register a subcommand.

        Args:

            description
                A one-line description to display in --help

            hidden
                When True don't display this command in --help
        """
        return self.milc.subcommand(description, hidden, **kwargs)

    def __enter__(self) -> Any:
        return self.milc.__enter__()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        return self.milc.__exit__(exc_type, exc_val, exc_tb)

    def add_spinner(self, name: str, spinner: Dict[str, Union[int, Sequence[str]]]) -> None:
        """Adds a new spinner to the list of spinners.

        A spinner is a dictionary with two keys:

            interval
                An integer that sets how long (in ms) to wait between frames.

            frames
                A list of frames for this spinner
        """
        return self.milc.add_spinner(name, spinner)

    def spinner(
        self,
        text: str,
        *args: Any,
        spinner: Optional[str] = None,
        animation: str = 'ellipsed',
        placement: str = 'left',
        color: str = 'blue',
        interval: int = -1,
        stream: Any = sys.stdout,
        enabled: bool = True,
        **kwargs: Any,
    ) -> Halo:
        """Create a spinner object for showing activity to the user.

        This uses halo <https://github.com/ManrajGrover/halo> behind the scenes, most of the arguments map to Halo objects 1:1.

        There are 3 basic ways to use this:

        * Instantiating a spinner and then using `.start()` and `.stop()` on your object.
        * Using a context manager (`with cli.spinner(...):`)
        * Decorate a function (`@cli.spinner(...)`)

        #### Instantiating a spinner

        ```python
        spinner = cli.spinner(text='Loading', spinner='dots')
        spinner.start()

        # Do something here

        spinner.stop()
        ```

        #### Using a context manager

        ```python
        with cli.spinner(text='Loading', spinner='dots'):
            # Do something here
        ```

        #### Decorate a function

        ```python
        @cli.spinner(text='Loading', spinner='dots')
        def long_running_function():
            # Do something here
        ```

        ### Arguments

            text
                The text to display next to the spinner. ANSI color strings
                (such as {fg_blue}) will be converted into ANSI escape
                sequences, and the ANSI reset sequence will be added to the
                end of the string.

                If *args or **kwargs are passed they will be used to
                %-format the text.

            spinner
                The name of the spinner to use. Available names are here:
                <https://raw.githubusercontent.com/sindresorhus/cli-spinners/dac4fc6571059bb9e9bc204711e9dfe8f72e5c6f/spinners.json>

            animation
                The animation to apply to the text if it doesn't fit the
                terminal. One of `ellipsed`, `bounce`, `marquee`.

            placement
                Which side of the text to display the spinner on. One of
                `left`, `right`.

            color
                Color of the spinner. One of `blue`, `grey`, `red`, `green`,
                `yellow`, `magenta`, `cyan`, `white`

            interval
                How long in ms to wait between frames. Defaults to the spinner interval (recommended.)

            stream
                Stream to write the output. Defaults to sys.stdout.

            enabled
                Enable or disable the spinner. Defaults to `True`.
        """
        return self.milc.spinner(text, *args, spinner=spinner, animation=animation, placement=placement, color=color, interval=interval, stream=stream, enabled=enabled, **kwargs)



import sys

from .common import check_assert, check_command, check_returncode


def test_custom_logger():
    result = check_command(sys.executable, 'custom_logger')
    check_returncode(result)
    check_assert(result, 'INFO:custom_logger:Hello Info World!' in result.stdout)
    check_assert(result, 'DEBUG:custom_logger:Hello Debug World!' in result.stdout)



"""Sometimes you need to ask the user a question. MILC provides basic functions for collecting and validating user input. You can find these in the `milc.questions` module.
"""
from getpass import getpass
from typing import Any, Callable, Optional, Sequence, TypeVar, Union, overload

from typing_extensions import Concatenate, ParamSpec

from milc import cli

from .ansi import format_ansi

T = TypeVar("T")
P = ParamSpec("P")


def _format_prompt(prompt: str, args: tuple, kwargs: dict) -> str:
    """Apply args or kwargs to a prompt string, or return the prompt unchanged.

    Args:
        prompt: The prompt string, optionally containing %-style format specifiers.
        args: Positional format arguments. Applied when non-empty.
        kwargs: Keyword format arguments. Applied when args is empty and kwargs is non-empty.

    Returns:
        The formatted prompt string, or the original prompt if neither args nor kwargs are provided.
    """
    if args:
        return prompt % args
    elif kwargs:
        return prompt % kwargs
    return prompt


def yesno(prompt: str, *args: Any, default: Optional[bool] = None, **kwargs: Any) -> bool:
    """Displays `prompt` to the user and gets a yes or no response.

    Returns `True` for a yes and `False` for a no.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like `cli.echo()`. |
    | default | Whether to default to a Yes or No when the user presses enter.<br><br>None- force the user to enter Y or N<br>True- Default to yes<br>False- Default to no |

    If you add `--yes` and `--no` arguments to your program the user can answer questions by passing command line flags.

    ```python
    @cli.argument('-y', '--yes', action='store_true', arg_only=True, help='Answer yes to all questions.')
    @cli.argument('-n', '--no', action='store_true', arg_only=True, help='Answer no to all questions.')
    ```
    """
    if args and kwargs:
        raise ValueError("You can't pass both args and kwargs!")

    # Check if we should return an answer without asking
    if 'no' in cli.args and cli.args.no:
        return False

    if 'yes' in cli.args and cli.args.yes:
        return True

    if not cli.interactive:
        return False

    # Format the prompt
    formatted_prompt = _format_prompt(prompt, args, kwargs)

    if default is None:
        formatted_prompt = formatted_prompt + ' [y/n] '
    elif default:
        formatted_prompt = formatted_prompt + ' [Y/n] '
    else:
        formatted_prompt = formatted_prompt + ' [y/N] '

    # Get input from the user
    while True:
        try:
            answer = input(format_ansi(formatted_prompt))
        except EOFError:
            return default if default is not None else False

        if not answer and default is not None:
            return default

        elif answer.lower() in ['y', 'yes']:
            return True

        elif answer.lower() in ['n', 'no']:
            return False


def password(
    prompt: str = 'Enter password:',
    *args: Any,
    confirm: bool = False,
    confirm_prompt: str = 'Confirm password:',
    confirm_limit: int = 3,
    validate: Optional[Callable[[str], bool]] = None,
    **kwargs: Any,
) -> Optional[str]:
    """Securely receive a password from the user. Returns the password or None.

    When running in non-interactive mode this will always return None. Otherwise it will return the confirmed password the user provides.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | confirm | Prompt the user to type the password again and make sure they match. |
    | confirm_prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | confirm_limit | Number of attempts to confirm before giving up. Default: 3 |
    | validate | This is an optional function that can be used to validate the password, EG to check complexity. It should return True or False and have the following signature:<br><br>`def function_name(answer):` |
    """
    if not cli.interactive:
        return None

    formatted_prompt = _format_prompt(prompt, args, kwargs)

    if formatted_prompt[-1] != ' ':
        formatted_prompt += ' '

    if confirm_prompt[-1] != ' ':
        confirm_prompt += ' '

    i = 0

    while confirm_limit == 0 or i < confirm_limit:
        try:
            pw = getpass(format_ansi(formatted_prompt))
        except EOFError:
            return None

        if pw:
            if validate is not None and not validate(pw):
                pass

            elif confirm:
                if getpass(format_ansi(_format_prompt(confirm_prompt, args, kwargs))) == pw:
                    return pw
                else:
                    cli.log.error('Passwords do not match!')

            else:
                return pw

            i += 1

    return None


def _cast_answer(answer_type: Callable[[str], T], answer: str) -> Optional[T]:
    """Attempt to convert answer to answer_type.
    """
    try:
        return answer_type(answer)
    except Exception as e:
        cli.log.error('Could not convert answer (%s) to type %s: %s', answer, getattr(answer_type, '__name__', repr(answer_type)), str(e))
        return None


@overload
def question(
    prompt: str,
    *args: Any,
    default: Optional[str] = ...,
    confirm: bool = ...,
    answer_type: None = ...,
    validate: Optional[Callable[Concatenate[str, P], bool]] = ...,
    **kwargs: Any,
) -> Optional[str]:
    ...


@overload
def question(
    prompt: str,
    *args: Any,
    default: Optional[str] = ...,
    confirm: bool = ...,
    answer_type: Callable[[str], T] = ...,
    validate: Optional[Callable[Concatenate[str, P], bool]] = ...,
    **kwargs: Any,
) -> Optional[T]:
    ...


# NOTE: can't have a default value on an argument whose type annotation is a TypeVar
# this means that `answer_type: Callable[[str], T] = str` gives a typing error
# see https://github.com/python/mypy/issues/3737
#
# due to this, we leave the default as `None`, while the actual implementation
# lives on a private function that receives all of its arguments from the public API.
# by doing this, the default value is "resolved" on callsite instead, making mypy happy.
#
# for better expresiveness, @overload variants are defined, to let the user know:
#   a) no `answer_type` provided: return str | None
#   b) `answer_type` converts str into T: return T | None
def question(
    prompt: str,
    *args: Any,
    default: Optional[str] = None,
    confirm: bool = False,
    answer_type: Optional[Callable[[str], T]] = None,
    validate: Optional[Callable[Concatenate[str, P], bool]] = None,
    **kwargs: Any,
) -> Union[str, T, None]:
    """Allow the user to type in a free-form string to answer.

    | Argument | Description |
    |----------|-------------|
    | prompt | The prompt to present to the user. Can include ANSI and format strings like milc's `cli.echo()`. |
    | default | The value to return when the user doesn't enter any value. Use None to prompt until they enter a value. |
    | confirm | Present the user with a confirmation dialog before accepting their answer. |
    | answer_type | Specify a type function for the answer. Will re-prompt the user if the function raises any errors. Common choices here include int, float, and decimal.Decimal. |
    | validate | This is an optional function that can be used to validate the answer. It should return True or False and have the following signature:<br><br>`def function_name(answer, *args, **kwargs):` |
    """
    return _question(
        prompt,
        *args,
        default=default,
        confirm=confirm,
        answer_type=answer_type or str,
        validate=validate,
        **kwargs,
    )


def _question(
    prompt: str,
    *args: Any,
    default: Optional[str],
    confirm: bool,
    answer_type: Callable[[str], T],
    validate: Optional[Callable[Concatenate[str, P], bool]],
    **kwargs: Any,
) -> Union[str, T, None]:
    if not cli.interactive:
        return default

    if default is not None:
        prompt = '%s [%s] ' % (prompt, default)
    elif prompt and prompt[-1] != ' ':
        prompt += ' '

    while True:
        try:
            answer = input(format_ansi(_format_prompt(prompt, args, kwargs)))
        except EOFError:
            return default

        if answer:
            if validate is not None and not validate(answer, *args, **kwargs):
                continue

            elif confirm:
                if yesno('Is the answer "%s" correct?', answer, default=True):
                    return _cast_answer(answer_type, answer)

            else:
                return _cast_answer(answer_type, answer)

        elif default is not None:
            return default


def choice(
    heading: str,
    options: Sequence[str],
    *args: Any,
    default: Optional[int] = None,
    confirm: bool = False,
    prompt: str = 'Please enter your choice: ',
    **kwargs: Any,
) -> Optional[str]:
    """Present the user with a list of options and let them select one.

    Users can enter either the number or the text of their choice. This will return the value of the item they choose, not the numerical index.

    | Argument | Description |
    |----------|-------------|
    | heading | The text to place above the list of options. |
    | options | A sequence of items to choose from. |
    | default | The index of the item to return when the user doesn't enter any value. Use None to prompt until they enter a value. |
    | confirm | When True present the user with a confirmation dialog before accepting their answer. |
    | prompt | The prompt to present to the user. Can include color and format strings like milc's `cli.echo()`. |

    Users can enter either the number or the text of their choice.

    !!! warning
        This will return the value of the item they choose, not the numerical index.
    """
    formatted_heading = _format_prompt(heading, args, kwargs)

    if not cli.interactive:
        if default is None:
            return None
        return options[default]

    if prompt[-1] != ' ':
        prompt += ' '

    if default is not None:
        prompt = '%s[%s] ' % (prompt, default + 1)

    while True:
        try:
            answer = _choice_get_answer(options, default, prompt, formatted_heading)
        except EOFError:
            return options[default] if default is not None else None

        if answer:
            if confirm and not yesno('Is the answer "%s" correct?', answer, default=True):
                continue

            return answer


def _choice_get_answer(
    options: Sequence[str],
    default: Optional[int],
    prompt: str,
    formatted_heading: str,
) -> Optional[str]:
    """Get an answer from the user for choice().
    """
    # Prompt for an answer.
    cli.echo(formatted_heading)

    for i, option in enumerate(options, 1):
        cli.echo('\t{fg_cyan}%d.{fg_reset} %s', i, option)

    answer = input(format_ansi(prompt))

    # If the user types in one of the options exactly use that
    if answer in options:
        return answer

    # Massage the answer into a valid integer
    if answer == '' and default is not None:
        answer_index = default
    elif answer.isnumeric():
        answer_index = int(answer) - 1
    else:
        cli.log.error('Invalid choice: %s', answer)
        return None

    # Validate the answer
    if answer_index >= len(options) or answer_index < 0:
        cli.log.error('Invalid choice: %s', answer_index + 1)
        return None

    # Return the answer they chose.
    return options[answer_index]



import sys

from .common import check_assert, check_command, check_returncode

# FIXME: Use something like this here: https://stackoverflow.com/questions/41542960/run-interactive-bash-with-popen-and-a-dedicated-tty-python
# def test_password_complexity():
#    result = check_command(sys.executable, 'passwd_complexity', '--interactive', input='aA1! \n')
#    check_returncode(result)
#    check_assert(result, 'Found 5 characters in the password!' in result.stdout)
#    check_assert(result, 'Found characters in 5 different character classes!' in result.stdout)


def test_password_complexity_no_input():
    result = check_command(sys.executable, 'passwd_complexity')
    check_returncode(result, 1)
    check_assert(result, 'No password provided!' in result.stdout)


# def test_password_confirm():
#    result = check_command(sys.executable, 'passwd_confirm', '--interactive', input='a\na')
#    check_returncode(result)
#    check_assert(result, 'Enter password:' in result.stdout)
#    check_assert(result, 'Confirm password:' in result.stdout)


def test_password_confirm_no_input():
    result = check_command(sys.executable, 'passwd_confirm')
    check_returncode(result, 1)



"""Unit tests for milc.configuration.
"""
import argparse

import milc.configuration

arg_parser = argparse.ArgumentParser(description='Minimal argparser.')


def test_get_argument_names_optional():
    """Make sure we can get optional names from get_argument_strings.
    """
    assert milc.configuration.get_argument_name(arg_parser, '-v', '--verbose') == 'verbose'


def test_get_argument_names_positional():
    """Make sure we can get positional names from get_argument_strings.
    """
    assert milc.configuration.get_argument_name(arg_parser, 'files') == 'files'


def test_get_argument_strings_optional():
    """Make sure we can get optional names from get_argument_strings.
    """
    assert milc.configuration.get_argument_strings(arg_parser, '-v', '--verbose') == ['-v', '--verbose']


def test_get_argument_strings_positional():
    """Make sure we can get positional names from get_argument_strings.
    """
    assert milc.configuration.get_argument_strings(arg_parser, 'files') == ['files']



import sys

from .common import check_assert, check_command, check_returncode


def test_sparkline():
    result = check_command(sys.executable, 'sparkline', '--no-color')
    check_returncode(result)
    check_assert(result, result.stdout == 'ℹ sparkline1: ▄▄▃▅ ▆▃█▂▁▁\nℹ sparkline2: ▁▁██\nℹ sparkline3: ▁▁▅▅██\n')



import sys

from .common import check_assert, check_command, check_returncode


def test_questions():
    result = check_command(sys.executable, 'questions')
    check_returncode(result)
    check_assert(result, 'User has chosen to stop.' in result.stdout)
    check_assert(result, 'User is stopping.' in result.stdout)
    check_assert(result, 'Interesting answer: None' in result.stdout)


def test_questions_interactive():
    result = check_command(sys.executable, 'questions', '--interactive', input='y\n2\nbecause\n')
    check_returncode(result)
    check_assert(result, 'User has chosen to continue.' in result.stdout)
    check_assert(result, 'User is not stopping.' in result.stdout)
    check_assert(result, 'Interesting answer: because' in result.stdout)


def test_questions_yes():
    result = check_command(sys.executable, 'questions', '--yes')
    check_returncode(result)
    check_assert(result, 'User has chosen to continue.' in result.stdout)
    check_assert(result, 'User is stopping.' in result.stdout)
    check_assert(result, 'Interesting answer: None' in result.stdout)


def test_questions_no():
    result = check_command(sys.executable, 'questions', '--no')
    check_returncode(result)
    check_assert(result, 'User has chosen to stop.' in result.stdout)
    check_assert(result, 'User is stopping.' in result.stdout)
    check_assert(result, 'Interesting answer: None' in result.stdout)



from typing import TYPE_CHECKING, Any, Hashable, List

if TYPE_CHECKING:
    from .milc import MILC

from .attrdict import AttrDict


class Configuration(AttrDict):
    """Represents the running configuration.

    This class never raises IndexError, instead it will return None if a
    section or option does not yet exist.
    """
    def __getitem__(self, key: Hashable) -> Any:
        """Returns a config section, creating it if it doesn't exist yet.
        """
        if key not in self._data:
            self._data[key] = ConfigurationSection(self)

        return self._data[key]


class ConfigurationSection(Configuration):
    def __init__(self, parent: AttrDict, *args: Any, **kwargs: Any) -> None:
        super(ConfigurationSection, self).__init__(*args, **kwargs)
        self._parent = parent

    def __getitem__(self, key: Hashable) -> Any:
        """Returns a config value, pulling from the `user` section as a fallback.
        This is called when the attribute is accessed either via the get method or through [ ] index.
        """
        if key in self._data and self._data.get(key) is not None:
            return self._data[key]

        elif key in self._parent.user:
            return self._parent.user[key]

        return None

    def __getattr__(self, key: str) -> Any:
        """Returns the config value from the `user` section.
        This is called when the attribute is accessed via dot notation but does not exist.
        """
        if key[0] != '_':
            return self.__getitem__(key)

        return None

    def __setattr__(self, key: str, value: Any) -> None:
        """Sets dictionary value when an attribute is set.
        """
        if key[0] != '_':
            self._data[key] = value
        else:
            object.__setattr__(self, key, value)


class SubparserWrapper(object):
    """Wrap subparsers so we can track what options the user passed.
    """

    # We type `cli` as Any instead of MILC to avoid a circular import
    def __init__(self, cli: Any, submodule: Any, subparser: Any) -> None:
        self.cli = cli
        self.submodule = submodule
        self.subparser = subparser
        self._arg_parser = self.cli._arg_parser

        for attr in dir(subparser):
            if not hasattr(self, attr):
                setattr(self, attr, getattr(subparser, attr))

    def completer(self, completer: Any) -> None:
        """Add an arpcomplete completer to this subcommand.
        """
        self.subparser.completer = completer

    def add_argument(self, *args: Any, **kwargs: Any) -> None:
        """Add an argument for this subcommand.

        This also stores the default for the argument in `self.cli.default_arguments`.
        """
        if kwargs.get('action') == 'store_boolean':
            # Store boolean will call us again with the enable/disable flag arguments
            handle_store_boolean(self, *args, **kwargs)

        else:
            completer = None

            if kwargs.get('completer'):
                completer = kwargs['completer']
                del kwargs['completer']

            self.cli.acquire_lock()
            argument_name = get_argument_name(self.cli._arg_parser, *args, **kwargs)

            if completer:
                self.subparser.add_argument(*args, **kwargs).completer = completer
            else:
                self.subparser.add_argument(*args, **kwargs)

            if kwargs.get('action') == 'store_false':
                self.cli._config_store_false.append(argument_name)

            if kwargs.get('action') == 'store_true':
                self.cli._config_store_true.append(argument_name)

            if self.submodule not in self.cli.default_arguments:
                self.cli.default_arguments[self.submodule] = {}

            self.cli.default_arguments[self.submodule][argument_name] = kwargs.get('default')
            self.cli.release_lock()


def get_argument_strings(arg_parser: Any, *args: Any, **kwargs: Any) -> List[str]:
    """Takes argparse arguments and returns a list of argument strings or positional names.
    """
    try:
        return arg_parser._get_optional_kwargs(*args, **kwargs)['option_strings']  # type: ignore[no-any-return]

    except ValueError:
        return [arg_parser._get_positional_kwargs(*args, **kwargs)['dest']]


def get_argument_name(arg_parser: Any, *args: Any, **kwargs: Any) -> Any:
    """Takes argparse arguments and returns the dest name.
    """
    try:
        return arg_parser._get_optional_kwargs(*args, **kwargs)['dest']

    except ValueError:
        return arg_parser._get_positional_kwargs(*args, **kwargs)['dest']


def handle_store_boolean(self: 'MILC | SubparserWrapper', *args: Any, **kwargs: Any) -> Any:
    """Does the add_argument for action='store_boolean'.
    """
    disabled_args = None
    disabled_kwargs = kwargs.copy()
    disabled_kwargs['action'] = 'store_false'
    disabled_kwargs['dest'] = get_argument_name(self._arg_parser, *args, **kwargs)
    disabled_kwargs['help'] = 'Disable ' + kwargs['help']
    kwargs['action'] = 'store_true'
    kwargs['help'] = 'Enable ' + kwargs['help']

    for flag in args:
        if flag[:2] == '--':
            disabled_args = ('--no-' + flag[2:],)
            break

    if disabled_args is None:
        raise ValueError("action='store_boolean' requires at least one long flag (e.g. '--flag')")

    self.add_argument(*args, **kwargs)
    self.add_argument(*disabled_args, **disabled_kwargs)

    return (args, kwargs, disabled_args, disabled_kwargs)



import os
import subprocess
import sys
from tempfile import NamedTemporaryFile

from .common import check_assert, check_command, check_returncode


def test_hello():
    result = check_command(sys.executable, 'hello')
    check_returncode(result)
    check_assert(result, 'Hello, World, from cli.log.info!' in result.stdout)
    check_assert(result, 'Hello, World, from cli.echo!' in result.stdout)


def test_hello_help():
    result = check_command(sys.executable, 'hello', '--help')
    check_returncode(result)
    check_assert(result, '[-n NAME]' in result.stdout)
    check_assert(result, '[--no-comma]' in result.stdout)


def test_hello_no_color():
    result = check_command(sys.executable, 'hello', '--no-color')
    check_returncode(result)
    if 'ℹ' in result.stdout:
        check_assert(result, result.stdout == 'ℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')
    else:
        check_assert(result, result.stdout == 'INFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')


def test_hello_no_color_no_unicode():
    result = check_command(sys.executable, 'hello', '--no-color', '--no-unicode')
    check_returncode(result)
    check_assert(result, result.stdout == 'INFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')


def test_hello_no_color_no_comma():
    result = check_command(sys.executable, 'hello', '--no-color', '--no-comma')
    check_returncode(result)
    if 'ℹ' in result.stdout:
        check_assert(result, result.stdout == 'ℹ Hello World, from cli.log.info!\nHello World, from cli.echo!\n')
    else:
        check_assert(result, result.stdout == 'INFO Hello World, from cli.log.info!\nHello World, from cli.echo!\n')


def test_hello_no_color_no_unicode_no_comma():
    result = check_command(sys.executable, 'hello', '--no-color', '--no-unicode', '--no-comma')
    check_returncode(result)
    check_assert(result, result.stdout == 'INFO Hello World, from cli.log.info!\nHello World, from cli.echo!\n')


def test_hello_no_color_log_file():
    log_file = NamedTemporaryFile(delete=False)
    result = check_command(sys.executable, 'hello', '--no-color', '--log-file', log_file.name)
    check_returncode(result)
    if 'ℹ' in result.stdout:
        check_assert(result, result.stdout == 'ℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')
    else:
        check_assert(result, result.stdout == 'INFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')
    log_file_contents = log_file.read().decode('utf-8')
    log_file.close()
    os.unlink(log_file.name)
    if 'Hello, World, from cli.log.info!' not in log_file_contents:
        print('Log File Contents:')
        print(log_file_contents)
        print()
        raise AssertionError


def test_hello_no_color_no_unicode_log_file():
    log_file = NamedTemporaryFile(delete=False)
    result = check_command(sys.executable, 'hello', '--no-color', '--no-unicode', '--log-file', log_file.name)
    check_returncode(result)
    check_assert(result, result.stdout == 'INFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')
    log_file_contents = log_file.read().decode('utf-8')
    log_file.close()
    os.unlink(log_file.name)
    if 'Hello, World, from cli.log.info!' not in log_file_contents:
        print('Log File Contents:')
        print(log_file_contents)
        print()
        raise AssertionError


def test_hello_no_color_verbose():
    result = check_command(sys.executable, 'hello', '--no-color', '-v')
    check_returncode(result)
    if '☐' in result.stdout:
        check_assert(result, result.stdout == '☐ You used -v you lucky person!\nℹ Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')
    else:
        check_assert(result, result.stdout == 'DEBUG You used -v you lucky person!\nINFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')


def test_hello_no_color_no_unicode_verbose():
    result = check_command(sys.executable, 'hello', '--no-color', '--no-unicode', '-v')
    check_returncode(result)
    check_assert(result, result.stdout == 'DEBUG You used -v you lucky person!\nINFO Hello, World, from cli.log.info!\nHello, World, from cli.echo!\n')


def test_hello_no_color_env():
    """Verify that setting NO_COLOR env var disables color output."""
    env = {**os.environ, 'NO_COLOR': '1'}
    result = subprocess.run([sys.executable, 'hello', '--no-unicode'], env={**env, 'PYTHONUTF8': '1'}, capture_output=True, text=True, encoding='utf-8')
    combined = result.stdout + result.stderr
    check_returncode(result)
    check_assert(result, '\x1b[' not in combined)
    check_assert(result, 'Hello, World, from cli.log.info!' in combined)



#!/usr/bin/env python3
# coding=utf-8
"""MILC - A CLI Framework

PYTHON_ARGCOMPLETE_OK

MILC is an opinionated framework for writing CLI apps. It optimizes for the
most common unix tool pattern- small tools that are run from the command
line and generally do not feature any user interaction while they run.

For more details see the MILC documentation:

    <https://github.com/clueboard/milc/tree/master/docs>
"""
__VERSION__ = '1.10.0'

import logging
import os
import sys
import warnings
from typing import Any, Optional

from .emoji import EMOJI_LOGLEVELS
from .milc import MILC
from .milc_interface import MILCInterface

if 'MILC_IGNORE_DEPRECATED' not in os.environ:
    for name in ('MILC_APP_NAME', 'MILC_APP_VERSION', 'MILC_APP_AUTHOR'):
        if name in os.environ:
            warnings.warn(f'Using environment variable {name} is deprecated and will not be supported in the future, please use cli.milc_options() instead.', stacklevel=2)

cli = MILCInterface()


def set_metadata(
    *,
    name: Optional[str] = None,
    author: Optional[str] = None,
    version: Optional[str] = None,
    logger: Optional[logging.Logger] = None,
) -> MILCInterface:
    """Set metadata about your program.

    Deprecated: Use `cli.milc_options()` instead.

    This allows you to set the application's name, version, and/or author
    before executing your entrypoint. You can also pass your own logger here
    if you like.

    It's best to run this only once, and it must be run before you call `cli()`.
    """
    warnings.warn("milc.set_metadata has been deprecated, please use cli.milc_options() instead.", stacklevel=2)
    cli.milc_options(name=name, author=author, version=version, logger=logger)

    return cli


# Extra stuff people can import
from ._sparkline import sparkline  # noqa



# coding=utf-8
import argparse
import logging
import os
import shlex
import subprocess
import sys
from configparser import RawConfigParser
from decimal import Decimal
from pathlib import Path
from platform import platform
from tempfile import NamedTemporaryFile
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Tuple, Type, TypeVar, Union

if TYPE_CHECKING:
    from argparse import _SubParsersAction

import threading

import argcomplete
import colorama
from halo import Halo
from platformdirs import user_config_dir
from spinners.spinners import Spinners
from typing_extensions import ParamSpec

from ._in_argv import _in_argv, _index_argv
from .ansi import MILCFormatter, ansi_colors, ansi_config, ansi_escape, format_ansi
from .attrdict import AttrDict
from .configuration import Configuration, SubparserWrapper, get_argument_name, get_argument_strings, handle_store_boolean

P = ParamSpec("P")
R = TypeVar("R")


class MILC(object):
    """MILC - An Opinionated Batteries Included Framework
    """
    def __init__(self, name: Optional[str] = None, author: Optional[str] = None, version: Optional[str] = None, logger: Optional[logging.Logger] = None) -> None:
        """Initialize the MILC object.
        """
        # Set some defaults
        if not name:
            name = os.environ.get('MILC_APP_NAME') or self.argv_name()

        if not version:
            version = os.environ.get('MILC_APP_VERSION', 'unknown')

        if not author:
            author = os.environ.get('MILC_APP_AUTHOR', name.upper())

        # Setup a lock for thread safety
        self._lock = threading.RLock()

        # Define some basic info
        self.acquire_lock()
        self.prog_name = name
        self.version = version
        self.author = author
        self._config_store_true: List[str] = []
        self._config_store_false: List[str] = []
        self._entrypoint: Callable[..., Any] = lambda _: None
        self._spinners: Dict[str, Dict[str, Union[int, Sequence[str]]]] = {}
        self._subcommand = None
        self._inside_context_manager = False
        self.ansi = ansi_colors
        self.arg_only: Dict[str, List[str]] = {}
        self.config_file = self.find_config_file()
        self.default_arguments: Dict[str, Dict[str, Optional[str]]] = {}
        self.platform = platform()
        self.interactive = sys.stdin.isatty()
        self.release_lock()
        self._deprecated_arguments: Dict[str, str] = {}
        self._deprecated_commands: Dict[str, str] = {}

        # Initialize all the things
        self.initialize_config()
        self.initialize_argparse()
        self.initialize_logging(logger)
        self.initialize_arguments()

    @property
    def config_dir(self) -> Path:
        return self.config_file.parent

    @property
    def description(self) -> Optional[str]:
        return self._arg_parser.description

    @description.setter
    def description(self, value: str) -> None:
        self._arg_parser.description = value

    @property
    def subcommand_name(self) -> Optional[str]:
        if self._subcommand is not None:
            name: str = self._subcommand.__name__
            return name

        return None

    def argv_name(self) -> str:
        """Returns the name of our program by examining argv.
        """
        app_name = sys.argv[0][:-3] if sys.argv[0].endswith('.py') else sys.argv[0]

        return os.path.split(app_name)[-1]

    def echo(self, text: str, *args: Any, **kwargs: Any) -> None:
        """Print colorized text to stdout.

        ANSI color strings (such as {fg_blue}) will be converted into ANSI
        escape sequences, and the ANSI reset sequence will be added to all
        strings.

        If *args or **kwargs are passed they will be used to %-format the strings.
        """
        if args and kwargs:
            raise ValueError('You can only specify *args or **kwargs, not both!')

        # FIXME: For V2, when we can break backwards compatibility, format_ansi first and then do the %-formatting
        if args:
            text = format_ansi(text % args)
        elif kwargs:
            text = format_ansi(text % kwargs)
        else:
            text = format_ansi(text)

        if not self.config.general.color:
            text = ansi_escape.sub('', text)

        print(text)

    def run(
        self,
        command: Sequence[str],
        capture_output: bool = True,
        combined_output: bool = False,
        text: bool = True,
        **kwargs: Any,
    ) -> Any:  # FIXME: In python 3.10 we can use subprocess.CompletedProcess[bytes | str] instead
        """Run a command using `subprocess.run`, but using some different defaults.

        Unlike subprocess.run you must supply a sequence of arguments. You can use `shlex.split()` to build this from a string.

        The **kwargs arguments get passed directly to `subprocess.run`.

        Args:
            command
                A sequence where the first item is the command to run, and any remaining items are arguments to pass.

            capture_output
                Set to False to have output written to the terminal instead of being available in the returned `subprocess.CompletedProcess` instance.

            combined_output
                When true STDERR will be written to STDOUT. Equivalent to the shell construct `2>&1`.

            text
                Set to False to disable encoding and get `bytes()` from `.stdout` and `.stderr`.
        """
        # Sanity Checking
        if isinstance(command, str):
            raise TypeError('`command` must be a non-text sequence such as list or tuple.')

        if not capture_output and combined_output:
            raise ValueError("Can't use capture_output=False and combined_output=True at the same time.")

        # On msys2/cygwin windows platforms you have to execute the command
        # through a subshell. As well, after execution stdin is broken so
        # things like milc.questions no longer work. We pass
        # `stdin=subprocess.DEVNULL` by default to prevent that.
        # On native Windows (no MSYSTEM env var), subprocess.run works directly.
        if 'windows' in self.platform.lower() and os.environ.get('MSYSTEM'):
            safecmd = ' '.join(map(shlex.quote, command))
            shell = os.environ.get('SHELL', 'bash')
            command = [shell, '-c', safecmd]

            if 'stdin' not in kwargs:
                kwargs['stdin'] = subprocess.DEVNULL

        # Argument Processing
        if capture_output:
            kwargs['stdout'] = subprocess.PIPE
            kwargs['stderr'] = subprocess.PIPE

        if combined_output:
            kwargs['stderr'] = subprocess.STDOUT

        if text:
            kwargs['universal_newlines'] = True

        # Run the command
        self.log.debug('Running command: %s', command)

        return subprocess.run(command, **kwargs)

    def initialize_argparse(self) -> None:
        """Prepare to process arguments from sys.argv.
        """
        kwargs = {
            'fromfile_prefix_chars': '@',
            'conflict_handler': 'resolve',
        }

        self.acquire_lock()

        self.subcommands: Dict[str, Any] = {}
        self._subparsers: Optional['_SubParsersAction[Any]'] = None
        self.argwarn = argcomplete.warn  # type: ignore[attr-defined]
        self.args = AttrDict()
        self.args_passed = AttrDict()
        self._arg_parser = argparse.ArgumentParser(**kwargs)  # type: ignore
        self.set_defaults = self._arg_parser.set_defaults

        self.release_lock()

    def print_help(self, *args: Any, **kwargs: Any) -> None:
        """Print a help message for the main program or subcommand, depending on context.
        """
        if self._subcommand:
            self.subcommands[self._subcommand.__name__].print_help(*args, **kwargs)
        else:
            self._arg_parser.print_help(*args, **kwargs)

    def print_usage(self, *args: Any, **kwargs: Any) -> None:
        """Print brief description of how the main program or subcommand is invoked, depending on context.
        """
        if self._subcommand:
            self.subcommands[self._subcommand.__name__].print_usage(*args, **kwargs)
        else:
            self._arg_parser.print_usage(*args, **kwargs)

    def log_deprecated_warning(self, item_type: str, name: str, reason: str) -> None:
        """Logs a warning with a custom message if an argument or command is deprecated.
        """
        self.log.warning("Warning: %s '%s' is deprecated:\n\t%s", item_type, name, reason)

    def add_argument(self, *args: Any, **kwargs: Any) -> None:
        """Wrapper to add arguments and track whether they were passed on the command line.
        """
        if 'action' in kwargs and kwargs['action'] == 'store_boolean':
            handle_store_boolean(self, *args, **kwargs)
            return

        arg_name = get_argument_name(self._arg_parser, *args, **kwargs)
        arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)

        # Handle tab completion
        completer = kwargs.get('completer')

        if 'completer' in kwargs:
            del kwargs['completer']

        self.acquire_lock()

        if completer:
            self._arg_parser.add_argument(*args, **kwargs).completer = completer  # type: ignore
        else:
            self._arg_parser.add_argument(*args, **kwargs)

        # Record the default for this argument
        if 'general' not in self.default_arguments:
            self.default_arguments['general'] = {}

        self.default_arguments['general'][arg_name] = kwargs.get('default')

        # Determine if it was passed on the command line
        if 'general' not in self.args_passed:
            self.args_passed['general'] = {}

        self.args_passed['general'][arg_name] = False

        for arg in arg_strings:
            if _in_argv(arg):
                self.args_passed['general'][arg_name] = True

        self.release_lock()

    def initialize_logging(self, logger: Optional[logging.Logger]) -> None:
        """Prepare the defaults for the logging infrastructure.
        """
        self.acquire_lock()

        if logger:
            self.milc_logger = False
            self.log = logger
        else:
            self.milc_logger = True
            self.log = logging.getLogger(self.__class__.__name__)
            self.log_file = None
            self.log_file_mode = 'a'
            self.log_file_handler: Optional[logging.FileHandler] = None
            self.log_print = True
            self.log_print_to = sys.stderr
            self.log_print_level = logging.INFO
            self.log_file_level = logging.INFO
            self.log_level = logging.INFO

            self.log.setLevel(logging.DEBUG)
            logging.root.setLevel(logging.DEBUG)

        self.release_lock()

        if self.milc_logger:
            self.add_argument('-v', '--verbose', action='store_true', help='Make the logging more verbose')
            self.add_argument('--datetime-fmt', default='%Y-%m-%d %H:%M:%S', help='Format string for datetimes')
            self.add_argument('--log-fmt', default='%(levelname)s %(message)s', help='Format string for printed log output')
            self.add_argument('--log-file-fmt', default='[%(levelname)s] [%(asctime)s] [file:%(pathname)s] [line:%(lineno)d] %(message)s', help='Format string for log file.')
            self.add_argument('--log-file-level', default='info', choices=['debug', 'info', 'warning', 'error', 'critical'], help='Logging level for log file.')
            self.add_argument('--log-file', help='File to write log messages to')
            self.add_argument('--color', action='store_boolean', default=ansi_config['color'], help='color in output')
            self.add_argument('--unicode', action='store_boolean', default=ansi_config['unicode'], help='unicode loglevels')

    def initialize_arguments(self) -> None:
        """Setup and add default arguments.
        """
        self.add_argument('-V', '--version', version=self.version, action='version', help='Display the version and exit')
        self.add_argument('--interactive', action='store_true', help='Force interactive mode even when stdout is not a tty.')
        self.add_argument('--config-file', help='The location for the configuration file')

        self.arg_only['config_file'] = ['general']

    def add_subparsers(self, title: str = 'Sub-commands', **kwargs: Any) -> None:
        if self._inside_context_manager:
            raise RuntimeError('You must run this before the with statement!')

        self.acquire_lock()
        self._subparsers = self._arg_parser.add_subparsers(title=title, dest='subparsers', **kwargs)
        self.release_lock()

    def acquire_lock(self, blocking: bool = True) -> bool:
        """Acquire the MILC lock for exclusive access to properties.
        """
        if self._lock:
            return self._lock.acquire(blocking)

        return True

    def release_lock(self) -> None:
        """Release the MILC lock.
        """
        if self._lock:
            self._lock.release()

    def find_config_file(self) -> Path:
        """Locate the config file.
        """
        config_file_index = _index_argv('--config-file')

        if config_file_index is not None:
            config_file_param = sys.argv[config_file_index]

            if '=' in config_file_param:
                # get the file name from the '=' assignment
                opt, config_file = config_file_param.split('=')

            else:
                # assume the file name is next space-sep arg
                if config_file_index + 1 >= len(sys.argv):
                    print('ERROR: --config-file requires a value.', file=sys.stderr)
                    sys.exit(1)
                config_file = sys.argv[config_file_index + 1]

            return Path(config_file).expanduser().resolve()

        filedir = user_config_dir(appname=self.prog_name, appauthor=self.author)
        filename = '%s.ini' % self.prog_name

        return Path(filedir, filename).resolve()

    def _handle_deprecated(self, arg_name: str, kwargs: Dict[str, Any]) -> None:
        """Called by self.argument: Mark an argument as deprecated, if necessary.
        """
        if 'deprecated' in kwargs:
            self._deprecated_arguments[arg_name] = kwargs['deprecated']
            if kwargs.get('help'):
                kwargs['help'] += f" [Deprecated]: {kwargs['deprecated']}"
            else:
                kwargs['help'] = f"[Deprecated]: {kwargs['deprecated']}"
            del kwargs['deprecated']

    def _handle_arg_parsing(self, config_name: str, arg_name: str, args: Sequence[Any], kwargs: Dict[str, Any]) -> None:
        """Called by self.argument: Parse this argument into the right datastructures.
        """
        arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)

        if kwargs.get('arg_only'):
            if arg_name not in self.arg_only:
                self.arg_only[arg_name] = []

            self.arg_only[arg_name].append(config_name)
            del kwargs['arg_only']
        else:
            if arg_name not in self.default_arguments:
                self.default_arguments[config_name] = {}

            self.default_arguments[config_name][arg_name] = kwargs.get('default')

            if self.config[config_name][arg_name] is None:
                self.config[config_name][arg_name] = kwargs.get('default')

            if config_name not in self.args_passed:
                self.args_passed[config_name] = {}

            self.args_passed[config_name][arg_name] = False

            for arg in arg_strings:
                if _in_argv(arg):
                    self.args_passed[config_name][arg_name] = True

    def argument(self, *args: Any, **kwargs: Any) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator to call self.add_argument or self.<subcommand>.add_argument.
        """
        if self._inside_context_manager:
            raise RuntimeError('You must run this before the with statement!')

        def argument_function(handler: Callable[P, R]) -> Callable[P, R]:
            config_name = getattr(handler, '__name__')
            subcommand_name = config_name.replace("_", "-")
            arg_name = get_argument_name(self._arg_parser, *args, **kwargs)

            self._handle_deprecated(arg_name, kwargs)
            self._handle_arg_parsing(config_name, arg_name, args, kwargs)

            if handler is self._entrypoint:
                self.add_argument(*args, **kwargs)

            elif subcommand_name in self.subcommands:
                self.subcommands[subcommand_name].add_argument(*args, **kwargs)

            else:
                raise RuntimeError('Decorated function is not entrypoint or subcommand!')

            return handler

        return argument_function

    def parse_args(self) -> None:
        """Parse the CLI args.
        """
        if self.args:
            self.log.debug('Warning: Arguments have already been parsed, ignoring duplicate attempt!')
            return

        argcomplete.autocomplete(self._arg_parser)

        self.acquire_lock()

        for key, value in vars(self._arg_parser.parse_args()).items():
            self.args[key] = value

        if 'entrypoint' in self.args:
            self._subcommand = self.args.entrypoint

        self.release_lock()

    def read_config_file(self) -> Tuple[Configuration, Configuration]:
        """Read in the configuration file and return Configuration objects for it and the config_source.
        """
        config = Configuration()
        config_source = Configuration()

        if self.config_file.exists():
            raw_config = RawConfigParser()
            raw_config.read(str(self.config_file))

            # Iterate over the config file options and write them into config
            for section in raw_config.sections():
                for option in raw_config.options(section):
                    value = raw_config.get(section, option)

                    # Coerce values into useful datatypes
                    if value.lower() in ['yes', 'true', 'on']:
                        value = True
                    elif value.lower() in ['no', 'false', 'off']:
                        value = False
                    elif value.lower() in ['none']:
                        continue
                    elif value.replace('.', '').isdigit():
                        if '.' in value:
                            value = Decimal(value)
                        else:
                            value = int(value)

                    config[section][option] = value
                    config_source[section][option] = 'config_file'

        return config, config_source

    def initialize_config(self) -> None:
        """Read in the configuration file and store it in self.config.
        """
        self.acquire_lock()
        self.config, self.config_source = self.read_config_file()
        self.release_lock()

    def merge_args_into_config(self) -> None:
        """Merge CLI arguments into self.config to create the runtime configuration.
        """
        self.acquire_lock()
        subcommand_name = self._subcommand.__name__ if self._subcommand else None

        for argument in self.args:
            if argument in ('subparsers', 'entrypoint'):
                continue

            # Find the argument's section. This is backwards because mypy is sticking its nose where it doesn't belong.
            if argument not in self.default_arguments['general']:
                section = subcommand_name
            else:
                section = 'general'

            if argument not in self.arg_only or section not in self.arg_only[argument]:
                # Determine the arg value and source
                arg_value = getattr(self.args, argument)

                # Merge this argument into self.config
                if self.args_passed[section][argument] or (argument in self._config_store_true and arg_value) or (argument in self._config_store_false and not arg_value):
                    self.config[section][argument] = arg_value
                    self.config_source[section][argument] = 'argument'
                elif self.config[section][argument] is None:
                    # Capture the default value
                    self.config[section][argument] = arg_value

        self.release_lock()

    def _save_config_file(self, config: AttrDict) -> None:
        """Write config to disk.
        """
        # Generate a sanitized version of our running configuration
        sane_config = RawConfigParser()
        for section_name, section in config.items():
            sane_config.add_section(section_name)
            for option_name, value in section.items():
                if self.config_source[section_name][option_name] == 'config_file' and value is not None:
                    sane_config.set(section_name, option_name, str(value))

        if not self.config_dir.exists():
            self.config_dir.mkdir(parents=True, exist_ok=True)

        # Write the config file atomically.
        self.acquire_lock()
        tmpfile_name = None
        try:
            with NamedTemporaryFile(mode='w', dir=str(self.config_dir), delete=False) as tmpfile:
                tmpfile_name = tmpfile.name
                sane_config.write(tmpfile)

            if os.path.getsize(tmpfile_name) > 0:
                os.replace(tmpfile_name, str(self.config_file))
            else:
                self.log.warning('Config file saving failed, not replacing %s with %s.', str(self.config_file), tmpfile_name)
        finally:
            try:
                if tmpfile_name and os.path.exists(tmpfile_name):
                    os.unlink(tmpfile_name)
            finally:
                self.release_lock()

    def write_config_option(self, section: str, option: Any) -> None:
        """Save a single config option to the config file.
        """
        if not self.config_file:
            self.log.warning('%s.config_file not set, not saving config!', self.__class__.__name__)
            return

        config, config_source = self.read_config_file()

        if section in config and option in config[section] and config[section][option] is None:
            del config[section][option]
        else:
            config[section][option] = str(self.config[section][option])

        self._save_config_file(config)

        # Housekeeping
        self.log.info('Wrote configuration to %s', shlex.quote(str(self.config_file)))

    def save_config(self) -> None:
        """Save the current configuration to the config file.
        """
        self.log.debug("Saving config file to '%s'", str(self.config_file))

        if not self.config_file:
            self.log.warning('%s.config_file file not set, not saving config!', self.__class__.__name__)
            return

        # Write config to disk
        self._save_config_file(self.config)
        self.log.info('Wrote configuration to %s', shlex.quote(str(self.config_file)))

    def check_deprecated(self) -> None:
        entry_name = getattr(self._entrypoint, '__name__')

        if entry_name in self._deprecated_commands:
            msg = self._deprecated_commands[entry_name]
            self.log_deprecated_warning('Entrypoint', entry_name, msg)

        if self._subcommand:
            name = self._subcommand.__name__.replace("_", "-")
            if name in self._deprecated_commands:
                msg = self._deprecated_commands[name]
                self.log_deprecated_warning('Subcommand', name, msg)

        deprecated_args_passed = [arg.split('=')[0].lstrip('-').replace('-', '_') for arg in sys.argv if arg.split('=')[0].lstrip('-').replace('-', '_') in self._deprecated_arguments]

        for arg in deprecated_args_passed:
            msg = self._deprecated_arguments[arg]
            self.log_deprecated_warning('Argument', arg, msg)

    def __call__(self) -> Any:
        """Execute the entrypoint function.
        """
        if not self._inside_context_manager:
            # If they didn't use the context manager use it ourselves
            with self:
                return self.__call__()

        self.check_deprecated()

        if self._subcommand:
            return self._subcommand(self)

        elif self._entrypoint is not None:
            return self._entrypoint(self)

        raise RuntimeError('No entrypoint provided!')

    def entrypoint(self, description: str, deprecated: Optional[str] = None) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator that marks the entrypoint used when a subcommand is not supplied.
        Args:
            description
                A one-line description to display in --help

            deprecated
                Deprecation message. When set the subcommand will marked as deprecated and this message will be displayed in the help output.
        """
        if self._inside_context_manager:
            raise RuntimeError('You must run this before cli()!')

        self.acquire_lock()
        self.description = description
        self.release_lock()

        def entrypoint_func(handler: Callable[P, R]) -> Callable[P, R]:
            self.acquire_lock()

            if deprecated:
                self._deprecated_commands[getattr(handler, '__name__')] = deprecated
                self.description = f'{self.description} [Deprecated]: {deprecated}'

            self._entrypoint = handler
            self.release_lock()

            return handler

        return entrypoint_func

    def add_subcommand(
        self,
        handler: Callable[P, R],
        description: str,
        hidden: bool = False,
        deprecated: Optional[str] = None,
        **kwargs: Any,
    ) -> Callable[P, R]:
        """Register a subcommand.

        Args:

            handler
                The function to execute for this subcommand.

            description
                A one-line description to display in --help

            hidden
                When True don't display this command in --help

            deprecated
                Deprecation message. When set the subcommand will be marked as deprecated
                and this message will be displayed in help output.
        """
        if self._inside_context_manager:
            raise RuntimeError('You must run this before the with statement!')

        if self._subparsers is None:
            self.add_subparsers(metavar="")

        name = getattr(handler, '__name__').replace("_", "-")

        if deprecated:
            self._deprecated_commands[name] = deprecated
            description += f' [Deprecated]: {deprecated}'

        self.acquire_lock()

        if not hidden and self._subparsers is not None:
            if self._subparsers.metavar:
                self._subparsers.metavar = "{%s,%s}" % (self._subparsers.metavar[1:-1], name)
            else:
                self._subparsers.metavar = "{%s}" % name
            kwargs['help'] = description

        assert self._subparsers is not None  # guaranteed by add_subparsers() call above
        self.subcommands[name] = SubparserWrapper(self, name, self._subparsers.add_parser(name, **kwargs))
        self.subcommands[name].set_defaults(entrypoint=handler)

        self.release_lock()

        return handler

    def subcommand(self, description: str, hidden: bool = False, **kwargs: Any) -> Callable[[Callable[P, R]], Callable[P, R]]:
        """Decorator to register a subcommand.

        Args:

            description
                A one-line description to display in --help

            hidden
                When True don't display this command in --help
        """
        def subcommand_function(handler: Callable[P, R]) -> Callable[P, R]:
            return self.add_subcommand(handler, description, hidden=hidden, **kwargs)

        return subcommand_function

    def setup_logging(self) -> None:
        """Called by __enter__() to setup the logging configuration.
        """
        if self.milc_logger:
            if len(logging.root.handlers) != 0:
                # MILC is the only thing that should have root log handlers
                logging.root.handlers = []

            self.acquire_lock()

            if self.config.general.verbose:
                self.log_print_level = logging.DEBUG

            ansi_config['color'] = self.config.general.color
            ansi_config['unicode'] = self.config.general.unicode

            self.log_file = self.config.general.log_file or self.log_file
            self.log_file_format = MILCFormatter(self.config.general.log_file_fmt, self.config.general.datetime_fmt)
            self.log_file_level = getattr(logging, self.config.general.log_file_level.upper())
            self.log_format = MILCFormatter(self.config.general.log_fmt, self.config.general.datetime_fmt)

            if self.log_file:
                self.log_file_handler = logging.FileHandler(self.log_file, self.log_file_mode)
                self.log_file_handler.setLevel(self.log_file_level)
                self.log_file_handler.setFormatter(self.log_file_format)
                logging.root.addHandler(self.log_file_handler)

            if self.log_print:
                self.log_print_handler = logging.StreamHandler(self.log_print_to)
                self.log_print_handler.setLevel(self.log_print_level)
                self.log_print_handler.setFormatter(self.log_format)
                logging.root.addHandler(self.log_print_handler)

            self.release_lock()

    def __enter__(self) -> Any:
        if self._inside_context_manager:
            self.log.debug('Warning: context manager was entered again. This usually means that self.__call__() was called before the with statement. You probably do not want to do that.')
            return

        self.acquire_lock()
        self._inside_context_manager = True
        self.release_lock()

        colorama.init()
        self.parse_args()
        self.merge_args_into_config()

        if self.config.general.interactive:
            self.interactive = True

        self.setup_logging()

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.acquire_lock()
        self._inside_context_manager = False
        self.release_lock()

        if exc_type is not None and not issubclass(exc_type, (SystemExit, KeyboardInterrupt)):
            self.log.error('%s: %s', exc_type.__name__, exc_val)
            sys.exit(255)

    def is_spinner(self, name: str) -> bool:
        """Returns true if name is a valid spinner.
        """
        return name in Spinners.__members__ or name in self._spinners

    def add_spinner(self, name: str, spinner: Dict[str, Union[int, Sequence[str]]]) -> None:
        """Adds a new spinner to the list of spinners.

        A spinner is a dictionary with two keys:

            interval
                An integer that sets how long (in ms) to wait between frames.

            frames
                A list of frames for this spinner
        """
        if self.is_spinner(name):
            raise ValueError(f'Spinner "{name}" already exists!')

        if not isinstance(spinner, dict):
            raise ValueError('The spinner must be a dictionary!')

        if 'interval' not in spinner or 'frames' not in spinner:
            raise ValueError('The spinner must have `interval` and `frames` keys!')

        self._spinners[name] = spinner

    def spinner(
        self,
        text: str,
        *args: Any,
        spinner: Optional[str] = None,
        animation: str = 'ellipsed',
        placement: str = 'left',
        color: str = 'blue',
        interval: int = -1,
        stream: Any = sys.stdout,
        enabled: bool = True,
        **kwargs: Any,
    ) -> Halo:
        """Create a spinner object for showing activity to the user.

        This uses halo <https://github.com/ManrajGrover/halo> behind the scenes, most of the arguments map to Halo objects 1:1.

        There are 3 basic ways to use this:

        * Instantiating a spinner and then using `.start()` and `.stop()` on your object.
        * Using a context manager (`with cli.spinner(...):`)
        * Decorate a function (`@cli.spinner(...)`)

        #### Instantiating a spinner

        ```python
        spinner = cli.spinner(text='Loading', spinner='dots')
        spinner.start()

        # Do something here

        spinner.stop()
        ```

        #### Using a context manager

        ```python
        with cli.spinner(text='Loading', spinner='dots'):
            # Do something here
        ```

        #### Decorate a function

        ```python
        @cli.spinner(text='Loading', spinner='dots')
        def long_running_function():
            # Do something here
        ```

        ### Arguments

            text
                The text to display next to the spinner. ANSI color strings
                (such as {fg_blue}) will be converted into ANSI escape
                sequences, and the ANSI reset sequence will be added to the
                end of the string.

                If *args or **kwargs are passed they will be used to
                %-format the text.

            spinner
                The name of the spinner to use. Available names are here:
                <https://raw.githubusercontent.com/sindresorhus/cli-spinners/dac4fc6571059bb9e9bc204711e9dfe8f72e5c6f/spinners.json>

            animation
                The animation to apply to the text if it doesn't fit the
                terminal. One of `ellipsed`, `bounce`, `marquee`.

            placement
                Which side of the text to display the spinner on. One of
                `left`, `right`.

            color
                Color of the spinner. One of `blue`, `grey`, `red`, `green`,
                `yellow`, `magenta`, `cyan`, `white`

            interval
                How long in ms to wait between frames. Defaults to the spinner interval (recommended.)

            stream
                Stream to write the output. Defaults to sys.stdout.

            enabled
                Enable or disable the spinner. Defaults to `True`.
        """
        spinner_name = spinner or 'line'  # FIXME: Grab one of the ascii spinners at random instead of line
        spinner_obj: Any = None

        if spinner in self._spinners:
            spinner_name = ''
            spinner_obj = self._spinners[spinner]

        if args:
            text = format_ansi(text % args)
        elif kwargs:
            text = format_ansi(text % kwargs)
        else:
            text = format_ansi(text)

        return Halo(
            text=text,
            spinner=spinner_name or spinner_obj,
            animation=None if animation == 'ellipsed' else animation,
            placement=placement,
            color=color,
            interval=interval,
        )



import sys

from .common import check_assert, check_command, check_returncode


def test_config_source():
    result = check_command(sys.executable, 'config_source')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, World, from None!\n')


def test_config_source_name():
    result = check_command(sys.executable, 'config_source', '--name', 'World')
    check_returncode(result)
    check_assert(result, result.stdout == 'Hello, World, from argument!\n')



from collections.abc import MutableMapping
from typing import Any, Iterator


class AttrDict(MutableMapping):
    """A dictionary that can also be accessed by attribute.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._data: dict = {}

    def __getitem__(self, key: Any) -> Any:
        return self._data[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: Any) -> None:
        del self._data[key]

    def __iter__(self) -> Iterator:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    def __contains__(self, key: Any) -> bool:
        return key in self._data

    def __getattr__(self, key: Any) -> Any:
        if key.startswith('_'):
            raise AttributeError(key)
        return self.__getitem__(key)


class SparseAttrDict(AttrDict):
    """A dictionary that can also be accessed by attribute.

    This class never raises IndexError, instead it will return None if a
    key does not yet exist.
    """
    def __getitem__(self, key: Any) -> Any:
        """Returns an item, creating it if it doesn't already exist
        """
        if key not in self._data:
            self._data[key] = None

        return self._data[key]
